import numpy as np
import util, random

class ClusterTree(object):
    """
    These are static cluster tree objects, you can construct them in one of two
    ways, either pass them a list of sets or a hierarchically structure list of
    lists.
    """
    
    def __init__(self, lists):
        """
        A ClusterTree maintains three representations of a hierarchical
        clustering. The first is a recursively structured list of lists. An
        example for a balanced binary hierarchy is:
        [[0,0,0,0,1,1,1,1],
         [[0,0,1,1], [[0,1], [[0]], [[0]]], [[0,1], [[0]], [[0]]]],
         [[0,0,1,1], [[0,1], [[0]], [[0]]], [[0,1], [[0]], [[0]]]]]         
        So the first element of the list defines the top level clusters and the
        subsequent elements are the recursively structured clusterings of the
        top clusters.
        
        The second structure is self.sets which is simply a flat list of all of
        the clusters (represented as sets)

        Finally there is self.DepthMatrix, which is a nxn symmetric matrix D
        where D[i,j] is the depth of the deepest cluster containing both objects
        i and j.
        """
        self.lists = None
        self.sets = None
        self.construct_from_lists(lists)
        self.depths()

    def construct_from_lists(self, lists):
        """
        Private method, invoked by __init__.
        """
        self.lists = lists
        self.n = len(self.lists[0])
        self._make_sets()
        self._order_clusters()

    def _make_sets(self):
        """
        From a ctree list, construct the corresponding sets, they are ordered
        recursively according to the hierarchy (i.e. all the sets corresponding
        to the first split are first, then the sets corresponding to the first
        split of the first subcluster are next and so on).
        """
        s1 = set(range(self.n))
        def make_set_recurse(subl, mapping):
            sets = []
            if len(subl) == 1:
                return []
            split = subl[0]
            inds = range(len(split))
            maxk = max(split)
            for i in range(maxk+1):
                sub_inds = [mapping[x] for x in range(len(inds)) if split[x] == i]
                sets.append(set(sub_inds))
            for i in range(maxk+1):
                sub_inds = [mapping[x] for x in range(len(inds)) if split[x] == i]
                sub_sets = make_set_recurse(subl[i+1], sub_inds)
                sets.extend(sub_sets)
            return sets
        sets = make_set_recurse(self.lists, range(self.n))
        self.sets = [s1]
        self.sets.extend(sets)
        

    def _order_clusters(self):
        """
        Given a cluster tree, find a permutation of the numbers 1...n that respects
        the relationships present in the cluster tree. A cluster tree can be
        obtained from matrices.make_arbitrary_hierarchical_block.
        """
        def order_cluster_recurse(ctree, depth=0):
            if len(ctree) == 1:
                return range(len(ctree[0]))
            split = ctree[0]
            inds = range(len(split))
            order = []
            for i in range(max(split)+1):
                sub_inds = [x for x in range(len(inds)) if split[x] == i]
                sub_order = order_cluster_recurse(ctree[i+1], depth=depth+1)
                sub_order = [sub_inds[x] for x in sub_order]
                order.extend(sub_order)
            return order
        self.order = order_cluster_recurse(self.lists)

    def misclustering(self, other):
        """
        This is like a comparator function, it returns the number of differences
        at each level of the heirarchy between this cluster tree and the other
        cluster tree.

        WARNING: this might be broken, seems to not terminate sometimes.
        """
        def miscluster_recurse(l1, l2, m1, m2):
            if len(l1) == 1 and len(l2) == 1:
                ## both lists are putting everything in one cluster, but they might be looking at different sets of objects.
                return len(set(m1).symmetric_difference(set(m2)))   
            else:
                s1s = [set([m1[x] for x in range(len(l1[0])) if l1[0][x] == k]) for k in range(max(l1[0])+1)]
                s2s = [set([m2[x] for x in range(len(l2[0])) if l2[0][x] == k]) for k in range(max(l2[0])+1)]
                (best, bestm) = util.misclustering(s1s, s2s)
                if len(l1) == 1 or len(l2) == 1:
                    ## stop recursing b/c one cluster tree didn't resolve lower levels, shouldn't punish for that.
                    return [best]
                sub_recs = []
                for i in range(len(bestm)):
#                     if bestm[i] == -1:
#                         ## then s1s[i] should not get mapped to anyone, but we should penalize it.
#                         sub_recs.append(miscluster_recurse(l1[i+1], [[]], [m1[x] for x in range(len(l1[0])) if l1[0][x] == i], []))
                    if bestm[i] != -1:
                        sub_recs.append(miscluster_recurse(l1[i+1], 
                                                           l2[bestm[i]+1],
                                                           [m1[x] for x in range(len(l1[0])) if l1[0][x] == i],
                                                           [m2[x] for x in range(len(l2[0])) if l2[0][x] == bestm[i]]))
                best = [best]
                best.extend(sub_recs)
                return best
        return miscluster_recurse(self.lists, other.lists, range(self.n), range(other.n))
                                           

    def depths(self):
        """
        Computes the DepthMatrix where the (i,j)th entry is the height of the
        deepest cluster that contains both points i and j.
        """
        def depth_recurse(curr, d):
            D = np.matrix(np.zeros((len(curr[0]), len(curr[0]))))
            ## just need to fill in the off diagonal, then recurse
            split = curr[0]
            if len(curr) == 1:
                D[curr,curr] = d*np.matrix(np.ones((len(curr), len(curr))))
                return D
            for i in range(max(split)+1):
                sub_inds = [x for x in range(len(split)) if split[x] == i]
                rest = [x for x in range(len(split)) if split[x] != i]
                D[np.ix_(sub_inds, rest)] = d*np.matrix(np.ones((len(sub_inds), len(rest))))
                D[np.ix_(rest, sub_inds)] = d*np.matrix(np.ones((len(rest), len(sub_inds))))
                D[np.ix_(sub_inds, sub_inds)] = depth_recurse(curr[i+1], d+1)
            return D
        D = depth_recurse(self.lists, 0)
        self.DepthMatrix = D

    def outlierfraction(self, truth, sampled = False):
        """
        Compute the fraction of outlier disagreements between this cluster tree
        and the true cluster tree.
        """
        if sampled:
            cands = random.sample(range(self.n), 100)
        else:
            cands = range(self.n)
        total = 0
        matching = 0
        for i in cands:
            for j in cands:
                for k in cands:
                    if (truth.DepthMatrix[i,j] >= max(truth.DepthMatrix[j,k], truth.DepthMatrix[i,k])):
                        total += 1
                        if (self.DepthMatrix[i,j] >= max(self.DepthMatrix[j,k], self.DepthMatrix[i,k])):
                            matching += 1
        return float(matching)/total
