import numpy as np
import random
import scipy.spatial.distance, scipy.cluster.hierarchy, scipy.sparse.linalg
from cluster_alg import *
import sys
from counting_matrix import *

"""
This module contains some basic algorithms for doing one-level
clustering. Others are around (for example in subsample_similarities,
subsample_objects, etc.). They can be used for k-way or as subroutines for
hierarchical algorithms.
"""
class ProjectionMethod(ClusterAlg):
    """
    Cluster the objects by fixing one row of the similarity matrix, and
    comparing every other row to it using cmp. For example, cmp can be the
    projection operator, in which case we project every row onto the subspace
    spanned by the first row and look at the norm. Another example is where cmp
    is the l2 distance. 

    The clusters are defined as the n/2 objects with the highest score. Note
    that this requires the number of clusters to be exactly 2 and the clusters
    to be perfectly balanced.
    """
    def __init__(self):
        super(ClusterAlg, self).__init__()
        self.name = "projection_method"

    def cluster(self, M, k=2, params={"cmp": lambda x,y: np.linalg.norm(x*(x.T*x)**(-1)*x.T*y, ord=2)}):
        if "cmp" not in params.keys():
            raise Exception("Projection Method Requires comparison parameter")
        if k != 2:
            raise Exception("ProjectionMethod does not support k-way clustering")
        M = self.preprocess(M)
        n = M.shape[0]
        x = M[0,:]
        scores = [params['cmp'](x, M[i,:]) for i in range(n)]
        s = sorted(zip(scores, range(n)), key=lambda x: x[0], reverse=False)
        g1 = [s[x][1] for x in range(n/2)]
        g2 = [s[x][1] for x in range(n/2, n)]
        sets = [set(g1), set(g2)]
        self.result = self.make_ctree(n, sets)

class Rank2Project(ClusterAlg):
    """
    This is McSherry's algorithm, without splitting the data. We first find the
    rank-2 approximation of the matrix (via eigendecomposition), and then we use
    projection method with l2 distance as the comparison.

    As with projection_method, this requires k=2, and perfectly balanced clusters
    """
    def __init__(self):
        super(ClusterAlg, self).__init__()
        self.name = "rank2_project"

    def cluster(self, M, k=2, params={}):
        if k != 2:
            raise Exception("Rank2Project does not support k-way clustering")
        M = self.preprocess(M)
        [v, S] = np.linalg.eigh(M)
        S = np.matrix(np.real(S))
        v = np.real(v)
        A = S[:,S.shape[1]-2:S.shape[1]]*np.diagflat(v[len(v)-2:len(v)])*S[:,S.shape[1]-2:S.shape[1]].T
        (g1, g2) = projection_method(A, cmp=lambda x,y: np.linalg.norm(x - y, ord=2))
        self.result = self.make_ctree(n, [set(g1), set(g2)])
        
class SpectralMethod(ClusterAlg):
    """
    This is the standard spectral graph clustering algorithm. We compute the
    graph laplacian, look at the top 2 eigenvectors and embed the data into this
    coordinate system. We cluster by choosing points that are far away from each
    other as seeds and mapping each point to its closest seed.
    """
    def __init__(self):
        super(ClusterAlg, self).__init__()
        self.name = "spectral"

    def cluster(self, M, k=2, params={}):
        M = self.preprocess(M)
        n = M.shape[0]
        D = np.diagflat(np.sum(M, axis=0))
        L = D - M
        [v, U] = np.linalg.eigh(L)
        idx = v.argsort()
        v = v[idx]
        U = U[:,idx]
        cs = []
        cs.append((U[0,0], U[0,1]))
        for i in range(k-1):
            dists = [min([(U[x,0] - cs[i][0])**2 + (U[x,1] - cs[i][1])**2 for i in range(len(cs))]) for x in range(n)]
            cand = np.argmax(dists)
            cs.append((U[cand,0], U[cand,1]))
        clusters = [set() for x in range(len(cs))]
        for i in range(n):
            dists = [(U[i,0] - cs[j][0])**2 + (U[i,1] - cs[j][1])**2 for j in range(len(cs))]
            clusters[np.argmin(dists)].add(i)
        self.result = self.make_ctree(n, clusters)

class SVDMethod(ClusterAlg):
    """
    This is a method for clustering using fewer observations. We subsample s(n)
    rows randomly and run the spectral method on A^T A.

    This requires k=2.
    """
    def __init__(self):
        super(ClusterAlg, self).__init__()
        self.name = "svd"

    def cluster(self, M, k=2, params={'s': lambda x: np.log(x)}):
        n = M.shape[0]
        if (type(M) == CountingMatrix):
            S = random.sample(range(n), int(param['s'](n)))
            A = np.matrix(M.row(S))
        else:
            S = random.sample(range(n), int(params['s'](n)))
            A = M[S,:]
        sub_alg = SpectralMethod()
        sub_alg.cluster(A*A.T, k)
        self.result = sub_alg.result


class KMeans(ClusterAlg):
    """
    Run k-means on the data matrix, if k is known we use k otherwise default to 2.
    """
    def __init__(self):
        super(ClusterAlg, self).__init__()
        self.name = "kmeans"

    def cluster(self, M, k=2, params={}):
        n = M.shape[0]
        if params['active']:
            if params['s'](n) > n:
                self.result = self.make_ctree(n, [set(range(n))])
                return
            if (type(M) == CountingMatrix):
                S = random.sample(range(n), int(params['s'](n)))
                A = np.matrix(M.row(S))
                A = A.T
            else:
                S = random.sample(range(n), int(params['s'](n)))
                A = M[S,:]
        else:
            S = range(n)
            A = np.matrix(self.preprocess(M))

        (codebook, blah) = scipy.cluster.vq.kmeans(A.T, k)
        (codes, blah) = scipy.cluster.vq.vq(A.T, codebook)
        sets = []
        for i in range(k):
            sets.append(set([x for x in range(n) if codes[x] == i]))
        self.result = self.make_ctree(n, sets)

