import numpy as np
import scipy.cluster.vq
import scipy.stats
import sys, random, copy
import cluster_tree, cluster_alg, util
from counting_matrix import *

"""
This is the algorithm that subsamples a set of objects, clusters them and uses
them to vote on the cluster assignments of the remaining objects. It is the
hierarchical version, so it recursively partitions the subclusters.
"""

class SOWithHeuristics(cluster_alg.ClusterAlg):
    """
    This is a flat, active spectral clustering algorithm with optional
    heuristics. It can take the following parameters (specified as the params
    dictionary to the cluster method): 
    
    'active' -- boolean -- determines whether to be active or not
    's' -- function -- # of objects to subsample when applied to problem size
    'drop' -- boolean -- drop low degree nodes?
    'stopping_cond' -- function -- when applied to problem size yields boolean
    saying if we should attempt to cluster or not.
    """
    def __init__(self):
        super(SOWithHeuristics, self).__init__()
        self.name = "so_heuristic"

    def check_params(self, params):
        if 'active' in params.keys() and params['active'] == True:
            if 's' not in params.keys():
                print "Warning: didn't see sampling parameter in params dictionary"
                params['s'] = lambda x: np.log2(x)
        else:
            params['active'] = False
        if 'drop' not in params.keys():
            print "Warning: didn't see drop parameter in params dictionary"
            params['drop'] = False
        if 'stopping_cond' not in params.keys():
            ## The stopping condition has to come from the hierarchical part of
            ## the algorithm, here it doesn't make a difference what it is
            ## because we only do one split
            print "Updating to default stopping condition"
            params['stopping_cond'] = lambda x: x <= 1

    def cluster(self, M, k=2, params={}):
        """
        1. Sample s(n) objects uniformly at random, observe the nxs(n) matrix
        2. look at the s(n) x s(n) matrix and compute degrees of each node
        3. Throw away any node whose degree is less than 2 st.devs. from the mean of the degree distribution. 
        4. Perform Spectral clustering (with heuristics) on the remainina sampled nodes.
        5. For each unclustered node, compute average similarity to each cluster. 
        6. If average similarity is too small, create a new cluster.
        """
        self.check_params(params)
        n = M.shape[0]
        if params['stopping_cond'](n):
            self.result = self.make_ctree(n, [set(range(n))])
            return
        if params['active']:
            if params['s'](n) > n:
                self.result = self.make_ctree(n, [set(range(n))])
                return
            if (type(M) == CountingMatrix):
                S = random.sample(range(n), int(params['s'](n)))
                A = M.row(S)
            else:
                S = random.sample(range(n), int(params['s'](n)))
                A = M[:,S]
            Asquare = np.matrix(A[S,:])
        else:
            S = range(n)
            Asquare = np.matrix(self.preprocess(M))

        ## Drop low degree nodes
        Degs = np.array(np.sum(Asquare, axis=0))[0]
        if params['drop']:
            sd = np.std(Degs)
            ave = np.mean(Degs)
            thresh = ave - 2*sd
            tokeep = []
            for i in range(len(Degs)):
                if Degs[i] >= ave-2*sd:
                    tokeep.append(i)
            Asquare = Asquare[np.ix_(tokeep, tokeep)]

        else:
            tokeep = range(len(Degs))
        seen = [S[i] for i in tokeep]
        SCalg = SpectralClusterWithHeuristics()
        SCalg.cluster(Asquare)
        sets = SCalg.result.sets[1:]
        set_idxs = []
        groups = []
        for s in sets:
            set_idxs.append([tokeep[i] for i in list(s)])
            groups.append(set([S[tokeep[i]] for i in list(s)]))
        if len(reduce(lambda x,y: x|y, groups, set([]))) == n:
            self.result = self.make_ctree(n, groups)
            return

        ## Can't do a line search over cut costs b/c we can't compute cut costs
        ## without looking at all of the similarities.  Instead, we're going to
        ## compute the distribution of within cluster similarity, using the
        ## subsampled nodes and if the max similarity is two st.devs below the
        ## mean then we will throw away this node
        distn = []
        for j in range(len(groups)):
            distn.extend([np.mean([util.get(M,i,k) for k in groups[j] if i != k]) for i in groups[j]])
        distn = [x for x in distn if x != np.nan]

        thresh = np.mean(distn) - 2*np.std(distn)
        set_groups = copy.deepcopy(groups)
        for i in range(n):
            if i not in seen:
                average_sims = [np.mean([util.get(M,i,k) for k in groups[j]]) for j in range(len(groups))]
                if params['make_new'] == False or np.max(average_sims) >= thresh:
                    set_groups[np.argmax(average_sims)].add(i)
                else:
                    groups.append(set([i]))
                    set_groups.append(set([i]))
        self.result = self.make_ctree(n, set_groups)
        return

class SpectralClusterWithHeuristics(cluster_alg.ClusterAlg):
    def cluster(self, M, k=2, params={}):
        M = self.preprocess(M)
        n = M.shape[0]
        [u, V] =  np.linalg.eig(np.diagflat(np.sum(M, axis=0)) - M)
        idx = u.argsort()   
        u = u[idx]
        V = V[:,idx]
#         if 'pick_k' in params.keys() and params['pick_k']:
#             fig = plt.figure()
#             ax1 = fig.add_subplot(111)
#             ax1.plot(u[0:20])
#             plt.show()
#             k = raw_input("Pick number of clusters k:")
#             try:
#                 k = int(k)
#             except ValueError:
#                 k = 2
#             plt.clf()
#         else:
        k = 2
        V = np.real(V)
        (codebook, blah) = scipy.cluster.vq.kmeans(V[:,0:k], k)
        (codes, blah) = scipy.cluster.vq.vq(V[:,0:k], codebook)
        sets = []
        for i in range(k):
            sets.append(set([x for x in range(n) if codes[x] == i]))
        self.result = self.make_ctree(n, sets)


