import numpy as np
import copy
import sys
from counting_matrix import *


"""
Several utility functions used throughout the clustering framework
"""

def dirichlet_groups(n, alpha):
    """
    Draw n data points from a stick breaking construction with parameter alpha.
    """
    groups = []
    group_counts = []
    for i in range(n):
        param = [float(group_counts[j])/(i+alpha) for j in range(len(group_counts))]
        param.append(float(alpha)/(i+alpha))
        g = np.random.multinomial(1, param)
        g = list(g).index(1)
        groups.append(g)
        if len(group_counts)-1 < g:
            group_counts.append(1)
        else:
            group_counts[g] += 1
    return groups


def binary_correctness(n, s1, s2):
    """
    Check if a clustering is correct. Assumes that range(0, n/2) is the first
    cluster and range(n/2, n) is the second cluster. s1, s2 are the cluster
    found by an algorithm.
    """
    t1 = set(range(n/2))
    t2 = set(range(n/2, n))
    
    if (s1 <= t1 and t1 <= s1 and s2 <= t2 and t2 <= s2):
        return True
    if (s1 <= t2 and t2 <= s1 and s2 <= t1 and t1 <= s2):
        return True
    return False


def binary_misclustering(n, s1, s2):
    """
    Compute the error of a binary misclustering, where the correct clustering is
    assumed to split the data in half (1, ..., n/2) and (n/2+1, ..., n). The
    binary misclustering is sum of the set differences between the correct and
    argument clusterings, for the best matching between the two, divided by the
    number of points.
    """
    t1 = set(range(n/2))
    t2 = set(range(n/2, n))
    g1 = t1.symmetric_difference(s1)
    g2 = t2.symmetric_difference(s2)
    g3 = t1.symmetric_difference(s2)
    g4 = t2.symmetric_difference(s1)
    return float(min((len(g1)+len(g2))/2, (len(g3)+len(g4))/2))/n

def largest_two_clusters(n, groups):
    """
    given a cluster hierarchy, extract the largest two groups, that hopefully
    form a partition of the set of nodes.
    """
    sizes = [(i, len(groups[i])) for i in range(len(groups))]
    ord = sorted(sizes, key=lambda x: x[1], reverse=True)
    g1 = None
    i = 0
    while ord[i][1] == n:
        i += 1
    g1 = groups[ord[i][0]]

    newgroups = [g for g in groups if len(g & g1) == 0]
    sizes = [(i, len(newgroups[i])) for i in range(len(newgroups))]
    ord = sorted(sizes, key=lambda x: x[1], reverse=True)
    g2 = newgroups[ord[0][0]]
    return (g1, g2)
    
def cut_cost(M, cut):
    """
    Given a graph cut, return the RatioCut objective for that cut.
    """
    if len(cut) == 1:
        return 0
    cost = 0
    allobj = set(range(M.shape[0]))
    for i in range(len(cut)):
        curr = list(cut[i])
        rest = list(allobj.difference(set(curr)))
        cost += float(np.sum(M[np.ix_(curr, rest)]))/len(cut[i])
    return cost

def misclustering(estimate, truth):
    """
    Given a set of clusters (flat) and the truth, find the lowest possible
    misclustering rate by looking at all mappings from estimate clusters to
    truth clusters. Also returns the best matching, which will be important for
    recursing down the hierarchy.
    """
    ## First enumerate all matchings. 
    def get_matchings(l1, l2):
        if len(l2) == 0 or len(l1) == 0:
            return [[]]
        matchings = []
        for i in l2:
            t1 = copy.copy(l1)
            t2 = copy.copy(l2)
            t1.pop()
            t2.remove(i)
            recs = get_matchings(t1, t2)
            for l in recs:
                ll = copy.copy(l)
                ll.append(i)
                matchings.append(ll)
        return matchings
    def inject_spaces(l, k):
        matchings = []
        if k == 0:
            return [l]
        if len(l) == 0:
            return [[-1 for i in range(k)]]
        for i in range(0, len(l)+1):
            sublist = l[0:i]
            sublist.append(-1)
            recs = inject_spaces(l[i:], k-1)
            for r in recs:
                c = copy.copy(sublist)
                c.extend(r)
                matchings.append(c)
        return matchings
    matchings = get_matchings(range(len(estimate)), range(len(truth)))
    if range(len(estimate)) > range(len(truth)):
        matching2s = []
        for matching in matchings:
            matching2s.extend(inject_spaces(matching, len(estimate) - len(truth)))
        matchings = matching2s

    ## now that we have all the matchings, we need to consider each one and look at the sum of the symmetric set differences
    best = -1
    bestm = None
    for m in matchings:
        score = float(sum([len(estimate[i].symmetric_difference(truth[m[i]])) for i in range(len(m)) if m[i] != -1]))
        score += float(sum([len(estimate[i]) for i in range(len(m)) if m[i] == -1]))
        score += float(sum([len(truth[i]) for i in range(len(truth)) if i not in m]))
        if score <= best or best == -1:
            best = score
            bestm = m
    return (best, bestm)
        
def get(M, i, j):
    """
    Short hand for handling both counting matrices and regular matrices
    easily. This function is useful to access any individual element of a
    matrix, regardless of type.
    """
    if type(M) == CountingMatrix:
        return M.get(i,j)
    else:
        return M[i,j]
