import numpy as np

class WordBag(object):
    """
    A bag of words. This is a wrapper around a python dictionary.
    """
    def __init__(self, words):
        """
        Given a list of words this constructs the dictionary object mapping words to counts.
        """
        self.map = {}
        for w in words:
            if w in self.map.keys():
                self.map[w] += 1
            else:
                self.map[w] = 1

    def add(self, word):
        """
        Add a word to the bag
        """
        if word in self.map.keys():
            self.map[word] += 1
        else:
            self.map[w] = 1

    def keys(self):
        """
        Get all the words in the bag
        """
        return self.map.keys()

    def get(self,key):
        """
        Get the counts associated with this word
        """
        return self.map[key]

    def items(self):
        """
        Get the (word, counts) pairs.
        """
        return self.map.items()

class TFIDFBag(object):
    """
    Another dictionary object, where words are mapped to their TFIDF
    scores. This requires a document in the context of a text corpus.
    """
    def __init__(self, wordbag, dfs, n):
        """
        Given document frequences and a wordbag, compute the Term Frequencies
        for this word bag and initialize a new TFIDFBag.
        """
        self.tfs = {}
        self.memoizedl2 = None
        for w in wordbag.keys():
            self.tfs[w] = wordbag.get(w)*np.log(float(n)/dfs[w])
        

    def l2(self):
        """
        Compute this bags l2 norm.
        """
        if self.memoizedl2 != None:
            return self.memoizedl2
        l2 = 0
        for (k,v) in self.tfs.items():
            l2 += v**2
        self.memoizedl2 = np.sqrt(l2)
        return self.memoizedl2

    def key_intersect(self, other):
        """
        Find the set of words that this bag shares in common with other.
        """
        return set(self.tfs.keys()).intersection(set(other.tfs.keys()))

    def cosinedist(self, other):
        """
        Compute the cosine distance between this bag and other.
        """
        return sum([self.tfs[w]*other.tfs[w] for w in self.key_intersect(other)])/(self.l2() * other.l2())

def compute_dfs(WBs):
    """
    Given a set of WordBags, compute the document frequencies.
    """
    dfs = {}
    for WB in WBs:
        for w in WB.keys():
            if w in dfs.keys():
                dfs[w] += 1
            else:
                dfs[w] = 1
    return dfs
